@@ -1,4 +1,4 @@
-# $Id: REST.pm 5 2006-07-22 21:10:23Z dmitri $
+# $Id: REST.pm 15 2006-07-25 18:02:34Z dmitri $
 # RT::Client::REST
 #
 # Dmitri Tikhonov <dtikhonov@vonage.com>
@@ -17,331 +17,19 @@
 #------------------------
 
 
-# We are going to throw exceptions, because we're cool like that.
-package RT::Client::REST::Exception;
-
-use strict;
-use warnings;
-
-use Error;
-
-use Exception::Class (
-    'RT::Client::REST::OddNumberOfArgumentsException'   => {
-        isa         => __PACKAGE__,
-        description => "This means that we wanted name/value pairs",
-    },
-
-    'RT::Client::REST::InvaildObjectTypeException'   => {
-        isa         => __PACKAGE__,
-        description => "Invalid object type was specified",
-    },
-
-    'RT::Client::REST::MalformedRTResponseException'    => {
-        isa         => __PACKAGE__,
-        description => "Malformed RT response received from server",
-    },
-
-    'RT::Client::REST::InvalidParameterValueException'  => {
-        isa         => __PACKAGE__,
-        description => "This happens when you feed me bad values",
-    },
-
-    'RT::Client::REST::RTException' => {
-        isa         => __PACKAGE__,
-        fields      => ['code'],
-        description => "RT server returned an error code",
-    },
-
-    'RT::Client::REST::ObjectNotFoundException' => {
-        isa         => 'RT::Client::REST::RTException',
-        description => 'One or more of the specified objects was not found',
-    },
-
-    'RT::Client::REST::CouldNotCreateObjectException' => {
-        isa         => 'RT::Client::REST::RTException',
-        description => 'Object could not be created',
-    },
-
-    'RT::Client::REST::AuthenticationFailureException'  => {
-        isa         => 'RT::Client::REST::RTException',
-        description => "Incorrect username or password",
-    },
-
-    'RT::Client::REST::UnknownRTException' => {
-        isa         => 'RT::Client::REST::RTException',
-        description => 'Some other RT error',
-    },
-
-    'RT::Client::REST::HTTPException'   => {
-        isa         => __PACKAGE__,
-        fields      => ['code'],
-        description => "Error in the underlying protocol (HTTP)",
-    },
-);
-
-sub _rt_content_to_exception {
-    my ($self, $content) = @_;
-
-    if ($content =~ /not found/) {
-        return 'RT::Client::REST::ObjectNotFoundException';
-    } elsif ($content =~ /not create/) {
-        return 'RT::Client::REST::CouldNotCreateObjectException';
-    } else {
-        return 'RT::Client::REST::UnknownRTException';
-    }
-}
-
-# Some mildly weird magic to fix up inheritance (see Exception::Class POD).
-{
-    no strict 'refs';
-    push @{__PACKAGE__ . '::ISA'}, 'Exception::Class::Base';
-    push @Exception::Class::Base::ISA, 'Error'
-        unless Exception::Class::Base->isa('Error');
-}
-
-
-# This package provides functions from RT::Interface::REST, in case
-# that module is not installed.  This one is simply copied from rt 3.4.5.
-package RT::Client::REST::Aux;
-
-use strict;
-use warnings;
-use Exporter;
-
-use vars qw(@EXPORT @ISA);
-
-@ISA = qw(Exporter);
-@EXPORT = qw(expand_list form_parse form_compose vpush vsplit);
-
-my $field = '(?:[a-zA-Z][a-zA-Z0-9_-]*|CF-[a-zA-Z0-9_ -]*)';
-
-sub expand_list {
-    my ($list) = @_;
-    my ($elt, @elts, %elts);
-
-    foreach $elt (split /,/, $list) {
-        if ($elt =~ /^(\d+)-(\d+)$/) { push @elts, ($1..$2) }
-        else                         { push @elts, $elt }
-    }
-
-    @elts{@elts}=();
-    return sort {$a<=>$b} keys %elts;
-}
-
-# Returns a reference to an array of parsed forms.
-sub form_parse {
-    my $state = 0;
-    my @forms = ();
-    my @lines = split /\n/, $_[0];
-    my ($c, $o, $k, $e) = ("", [], {}, "");
-
-    LINE:
-    while (@lines) {
-        my $line = shift @lines;
-
-        next LINE if $line eq '';
-
-        if ($line eq '--') {
-            # We reached the end of one form. We'll ignore it if it was
-            # empty, and store it otherwise, errors and all.
-            if ($e || $c || @$o) {
-                push @forms, [ $c, $o, $k, $e ];
-                $c = ""; $o = []; $k = {}; $e = "";
-            }
-            $state = 0;
-        }
-        elsif ($state != -1) {
-            if ($state == 0 && $line =~ /^#/) {
-                # Read an optional block of comments (only) at the start
-                # of the form.
-                $state = 1;
-                $c = $line;
-                while (@lines && $lines[0] =~ /^#/) {
-                    $c .= "\n".shift @lines;
-                }
-                $c .= "\n";
-            }
-            elsif ($state <= 1 && $line =~ /^($field):(?:\s+(.*))?$/) {
-                # Read a field: value specification.
-                my $f  = $1;
-                my @v  = ($2 || ());
-
-                # Read continuation lines, if any.
-                while (@lines && ($lines[0] eq '' || $lines[0] =~ /^\s+/)) {
-                    push @v, shift @lines;
-                }
-                pop @v while (@v && $v[-1] eq '');
-
-                # Strip longest common leading indent from text.
-                my ($ws, $ls) = ("");
-                foreach $ls (map {/^(\s+)/} @v[1..$#v]) {
-                    $ws = $ls if (!$ws || length($ls) < length($ws));
-                }
-                s/^$ws// foreach @v;
-
-                push(@$o, $f) unless exists $k->{$f};
-                vpush($k, $f, join("\n", @v));
-
-                $state = 1;
-            }
-            elsif ($line !~ /^#/) {
-                # We've found a syntax error, so we'll reconstruct the
-                # form parsed thus far, and add an error marker. (>>)
-                $state = -1;
-                $e = form_compose([[ "", $o, $k, "" ]]);
-                $e.= $line =~ /^>>/ ? "$line\n" : ">> $line\n";
-            }
-        }
-        else {
-            # We saw a syntax error earlier, so we'll accumulate the
-            # contents of this form until the end.
-            $e .= "$line\n";
-        }
-    }
-    push(@forms, [ $c, $o, $k, $e ]) if ($e || $c || @$o);
-
-    my $l;
-    foreach $l (keys %$k) {
-        $k->{$l} = vsplit($k->{$l}) if (ref $k->{$l} eq 'ARRAY');
-    }
-
-    return \@forms;
-}
-
-# Returns text representing a set of forms.
-sub form_compose {
-    my ($forms) = @_;
-    my (@text, $form);
-
-    foreach $form (@$forms) {
-        my ($c, $o, $k, $e) = @$form;
-        my $text = "";
-
-        if ($c) {
-            $c =~ s/\n*$/\n/;
-            $text = "$c\n";
-        }
-        if ($e) {
-            $text .= $e;
-        }
-        elsif ($o) {
-            my (@lines, $key);
-
-            foreach $key (@$o) {
-                my ($line, $sp, $v);
-                my @values = (ref $k->{$key} eq 'ARRAY') ?
-                               @{ $k->{$key} } :
-                                  $k->{$key};
-
-                $sp = " "x(length("$key: "));
-                $sp = " "x4 if length($sp) > 16;
-
-                foreach $v (@values) {
-                    if ($v =~ /\n/) {
-                        $v =~ s/^/$sp/gm;
-                        $v =~ s/^$sp//;
-
-                        if ($line) {
-                            push @lines, "$line\n\n";
-                            $line = "";
-                        }
-                        elsif (@lines && $lines[-1] !~ /\n\n$/) {
-                            $lines[-1] .= "\n";
-                        }
-                        push @lines, "$key: $v\n\n";
-                    }
-                    elsif ($line &&
-                           length($line)+length($v)-rindex($line, "\n") >= 70)
-                    {
-                        $line .= ",\n$sp$v";
-                    }
-                    else {
-                        $line = $line ? "$line, $v" : "$key: $v";
-                    }
-                }
-
-                $line = "$key:" unless @values;
-                if ($line) {
-                    if ($line =~ /\n/) {
-                        if (@lines && $lines[-1] !~ /\n\n$/) {
-                            $lines[-1] .= "\n";
-                        }
-                        $line .= "\n";
-                    }
-                    push @lines, "$line\n";
-                }
-            }
-
-            $text .= join "", @lines;
-        }
-        else {
-            chomp $text;
-        }
-        push @text, $text;
-    }
-
-    return join "\n--\n\n", @text;
-}
-
-# Add a value to a (possibly multi-valued) hash key.
-sub vpush {
-    my ($hash, $key, $val) = @_;
-    my @val = ref $val eq 'ARRAY' ? @$val : $val;
-
-    if (exists $hash->{$key}) {
-        unless (ref $hash->{$key} eq 'ARRAY') {
-            my @v = $hash->{$key} ne '' ? $hash->{$key} : ();
-            $hash->{$key} = \@v;
-        }
-        push @{ $hash->{$key} }, @val;
-    }
-    else {
-        $hash->{$key} = $val;
-    }
-}
-
-# "Normalise" a hash key that's known to be multi-valued.
-sub vsplit {
-    my ($val) = @_;
-    my ($line, $word, @words);
-
-    foreach $line (map {split /\n/} (ref $val eq 'ARRAY') ? @$val : $val)
-    {
-        # XXX: This should become a real parser, à la Text::ParseWords.
-        $line =~ s/^\s+//;
-        $line =~ s/\s+$//;
-        push @words, split /\s*,\s*/, $line;
-    }
-
-    return \@words;
-}
-
-# Now for the actual package.
 package RT::Client::REST;
 
 use strict;
 use warnings;
 
 use vars qw/$VERSION/;
-$VERSION = '0.11';
-
-# If RT::Interface::REST, use it.  Otherwise, use local copy.
-eval { require RT::Interface::REST };
-
-# This exports the following methods:
-# expand_list form_parse form_compose vpush vsplit
-if ($@) {
-    for my $function (qw(expand_list form_parse form_compose vpush vsplit)) {
-        no strict 'refs';
-        *{$function} = \&{'RT::Client::REST::Aux::' . $function};
-    }
-} else {
-    eval "use RT::Interface::REST;";
-}
+$VERSION = '0.12';
 
 use LWP;
 use HTTP::Cookies;
 use HTTP::Request::Common;
+use RT::Client::REST::Exception;
+use RT::Client::REST::Forms;
 
 # Generate accessors/mutators
 for my $method (qw(username password server cookie)) {
@@ -440,22 +128,26 @@ sub comment {
     my $ticket_id = $self->_valid_numeric_ticket_id(delete($opts{ticket_id}));
     my $msg = $self->_valid_comment_message(delete($opts{message}));
 
-    my $text = form_compose([[
-        '',
-        [ "Ticket", "Action", "Cc", "Bcc", "Attachment", "TimeWorked", "Text" ],
-        {
-            Ticket      => $ticket_id,
-            Action      => $action,
-            Cc          => [],
-            Bcc         => [],
-            Attachment  => [],
-            TimeWorked  => '',
-            Text        => $msg,
-            Status      => '',
-        },
-    ]]);
+    my @objects = ("Ticket", "Action", "Text");
+    my %values  = (
+        Ticket      => $ticket_id,
+        Action      => $action,
+        Text        => $msg,
+    );
 
-    $self->_submit("ticket/comment/$ticket_id", {
+    if (exists($opts{cc})) {
+        push @objects, "Cc";
+        $values{Cc} = delete($opts{cc});
+    }
+
+    if (exists($opts{bcc})) {
+        push @objects, "Bcc";
+        $values{Bcc} = delete($opts{bcc});
+    }
+
+    my $text = form_compose([[ '', \@objects, \%values, ]]);
+
+    $self->_submit("ticket/$ticket_id/comment", {
         content => $text,
     });
 
@@ -594,10 +286,7 @@ sub _submit {
                     #$session->delete;
                     #return submit(@_) unless $uri eq "$REST/logout";
                 }
-            }
-            # Conflicts should be dealt with by the handler and user.
-            # For anything else, we just die.
-            elsif ($res->code != 409) {
+            } else {
                 RT::Client::REST::Exception->_rt_content_to_exception(
                     $res->content)
                 ->throw(
@@ -800,13 +489,22 @@ Create a new object of type B<$type> and set initial parameters to B<%params>.
 Returns numeric ID of the new object.  If numeric ID cannot be parsed from
 the response, B<RT::Client::REST::MalformedRTResponseException> is thrown.
 
-=item comment (ticket_id => $id, message => "This is a comment")
+=item comment (ticket_id => $id, message => $message, %opts)
 
 Comment on a ticket with ID B<$id>.
+Optionally takes arguments B<cc> and B<bcc> which are references to lists
+of e-mail addresses:
 
-=item correspond (ticket_id => $id, message => "This is a comment")
+  $rt->comment(
+    ticket_id   => 5,
+    message     => "Wild thing, you make my heart sing",
+    cc          => [qw(dmitri@localhost some@otherdude.com)],
+  );
 
-Add correspondence to ticket ID B<$id>.
+=item correspond (ticket_id => $id, message => $message, %opts)
+
+Add correspondence to ticket ID B<$id>.  Takes optional B<cc> and
+B<bcc> parameters (see C<comment> above).
 
 =item merge_tickets (src => $id1, dst => $id2)
 
@@ -979,4 +677,4 @@ was later converted to this module by Dmitri Tikhonov <dtikhonov@vonage.com>
 
 Since original rt is licensed under GPL, so is this module.
 
-=end
+=cut
